from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Coroutine
from dataclasses import dataclass
from datetime import datetime
from sqlite3 import Blob
from typing import Any

from types_aiobotocore_s3.type_defs import (
    CompletedPartTypeDef,
)

MP_THRESHOLD = 25000000
MP_CHUNK_SIZE = 25000000


class EmptyMountResponseError(Exception):
    pass


@dataclass
class BlobMeta:
    bucket: str
    key: str
    content_type: Any
    size: int
    version: Any
    update_time: datetime | None


class StorageClient(ABC):
    @abstractmethod
    async def head(self, bucket_name: str, key: str) -> BlobMeta:
        pass

    @abstractmethod
    async def get_blob_bytes(
        self,
        bucket_name: str,
        key: str,
        start: int | None = None,
        end: int | None = None,
    ) -> bytes:
        pass

    @abstractmethod
    def list_keys(
        self,
        bucket_name: str,
        prefix: str | None = None,
        delimiter: str = "/",
    ) -> AsyncGenerator[str, Any]:
        pass

    @abstractmethod
    async def put_blob(
        self,
        bucket_name: str,
        key: str,
        body: bytes = b"",
        content_type: str | None = None,
        acl: str = "bucket-owner-full-control",
    ) -> BlobMeta:
        pass

    @abstractmethod
    async def copy_blob(
        self,
        src_key: str,
        src_bucket_name: str,
        dest_key: str,
        dest_bucket_name: str,
    ) -> None:
        pass

    @abstractmethod
    async def copy_blob_multipart(
        self,
        src_key: str,
        src_bucket_name: str,
        dest_key: str,
        dest_bucket_name: str,
    ) -> None:
        pass

    @abstractmethod
    async def delete_blob(
        self,
        bucket_name: str,
        key: str,
    ) -> None:
        pass

    @abstractmethod
    async def generate_signed_download_url(
        self,
        bucket_name: str,
        key: str | None,
        content_disposition: str | None = None,
        content_type: str | None = None,
    ) -> str:
        pass

    @abstractmethod
    async def generate_signed_part_upload_url(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
    ) -> str:
        pass

    @abstractmethod
    async def multipart_initiate_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        content_type: str,
        acl: str = "storageAdmin",
    ) -> str:
        pass

    @abstractmethod
    async def multipart_upload_part(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
        data: bytes,
    ) -> str:
        pass

    @abstractmethod
    async def multipart_complete_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        parts: list[CompletedPartTypeDef],
    ) -> str:
        pass



from collections.abc import AsyncGenerator
from typing import Any

from google.api_core.exceptions import (
    NotFound,
)

# from latch_o11y.o11y import trace_function_with_span
# from opentelemetry.trace import Span, get_tracer
from types_aiobotocore_s3.type_defs import (
    CompletedPartTypeDef,
)

from latch_cloud_clients.gcp.storage import Object

from ..utils import BlobMeta, EmptyMountResponseError, StorageClient
from .client_pool import gcp_pool


class GCPStorageClient(StorageClient):
    async def head(self, bucket_name: str, key: str) -> BlobMeta:
        # s.set_attributes({"bucket_name": bucket_name, "key": key})
        async with gcp_pool.gcp_client() as gcp:
            client = gcp.storage_client()
            blob = await client.get_blob_meta(bucket_name, key)

            if blob is None:
                blobs = []
                it = client.list_blobs(
                    bucket_name,
                    prefix=key.rstrip("/"),
                    delimiter="/",
                    max_items=1,
                    include_trailing_delimeter=True,
                    include_folders_as_prefixes=True,
                )
                async for blob in it:
                    blobs.append(blob)

                if len(blobs) > 0:
                    blob = blobs[0]
                else:
                    if len(it.prefixes) == 0:
                        raise EmptyMountResponseError()

                    if key.rstrip("/") not in [p.rstrip("/") for p in it.prefixes]:
                        raise EmptyMountResponseError()

                    blob = Object(
                        id=key,
                        bucket=bucket_name,
                        name=key,
                        content_type="dir",
                        media_link=None,
                        size=0,
                        generation=None,
                        update_time=None,
                        creation_time=None,
                    )

            blob = BlobMeta(
                bucket=bucket_name,
                key=key,
                content_type=blob.content_type,
                size=blob.size,
                version=blob.generation,
                update_time=blob.update_time,
            )

            return blob

    async def get_blob_bytes(
        self,
        bucket_name: str,
        key: str,
        start: int | None = None,
        end: int | None = None,
    ) -> bytes:
        # s.set_attributes({"bucket_name": bucket_name, "key": key})
        async with gcp_pool.gcp_client() as gcp:
            client = gcp.storage_client()

            try:
                return await client.get_blob_bytes(
                    bucket_name, key, start=start, end=end
                )
            except NotFound as e:
                raise EmptyMountResponseError() from e

    async def list_keys(
        self,
        bucket_name: str,
        prefix: str | None = None,
        delimiter: str = "/",
    ) -> AsyncGenerator[str, Any]:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "prefix": str(prefix),
        #         "delimeter": delimiter,
        #         "page_size": page_size,
        #     }
        # )
        async with gcp_pool.gcp_client() as gcp:
            client = gcp.storage_client()

            iterator = client.list_blobs(
                bucket_name,
                prefix=prefix,
                delimiter=delimiter,
            )

            async for obj in iterator:
                # don't add the prefix itself to the list
                if prefix is None or obj.name.rstrip("/") != prefix.rstrip("/"):
                    yield obj.name

            # all the prefixes are present after executing iterator above
            blob_prefixes = set(iterator.prefixes)

            for p in blob_prefixes:
                if prefix is None or p.rstrip("/") != prefix.rstrip("/"):
                    yield p

            return

    async def put_blob(
        self,
        bucket_name: str,
        key: str,
        body: bytes = b"",
        content_type: str | None = None,
        acl: str = "bucket-owner-full-control",
    ) -> BlobMeta:
        # s.set_attributes({"bucket_name": bucket_name, "key": key})

        async with gcp_pool.gcp_client() as gcp:
            client = gcp.storage_client()

            content_type = content_type if content_type else "text/plain"

            obj = await client.put_blob(
                bucket_name, key, body, content_type=content_type
            )

            return BlobMeta(
                bucket=bucket_name,
                key=key,
                content_type=content_type,
                size=len(body),
                version=obj.generation,
                update_time=obj.update_time,
            )

    async def copy_blob(
        self,
        src_key: str,
        src_bucket_name: str,
        dest_key: str,
        dest_bucket_name: str,
    ) -> None:
        # s.set_attributes(
        #     {
        #         "src_key": src_key,
        #         "src_bucket_name": src_bucket_name,
        #         "dest_key": dest_key,
        #         "dest_bucket_name": dest_bucket_name,
        #     }
        # )

        async with gcp_pool.gcp_client() as gcp:
            client = gcp.storage_client()

            try:
                await client.copy_blob(
                    src_bucket_name=src_bucket_name,
                    src_bucket_key=src_key,
                    dest_bucket_name=dest_bucket_name,
                    dest_bucket_key=dest_key,
                )
            except NotFound as e:
                # s.record_exception(e)
                raise EmptyMountResponseError from e

    async def copy_blob_multipart(
        self,
        src_key: str,
        src_bucket: str,
        dest_key: str,
        dest_bucket: str,
    ) -> None:
        # s.set_attributes(
        #     {
        #         "src_key": src_key,
        #         "src_bucket": src_bucket,
        #         "dest_key": dest_key,
        #         "dest_bucket": dest_bucket,
        #     }
        # )
        await self.copy_blob(src_key, src_bucket, dest_key, dest_bucket)

    async def delete_blob(
        self,
        bucket_name: str,
        key: str,
    ) -> None:
        # s.set_attributes({"bucket_name": bucket_name, "key": key})
        async with gcp_pool.gcp_client() as gcp:
            await gcp.storage_client().delete_blob(bucket_name, key)

    async def generate_signed_download_url(
        self,
        bucket_name: str,
        key: str | None,
        content_disposition: str | None = None,
        content_type: str | None = None,
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "key": str(key),
        #         "content_disposition": str(content_disposition),
        #         "content_type": str(content_type),
        #     }
        # )
        async with gcp_pool.gcp_client() as gcp:
            url = gcp.storage_client().get_signed_download_url(
                bucket_name=bucket_name,
                key=key,
                content_disposition=content_disposition,
                content_type=content_type,
            )

            return url

    async def generate_signed_part_upload_url(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "bucket_key": bucket_key,
        #         "idx": part_number,
        #     }
        # )

        async with gcp_pool.gcp_client() as gcp:
            return await gcp.storage_client().get_signed_part_upload_url(
                bucket_name=bucket_name,
                bucket_key=bucket_key,
                upload_id=upload_id,
                part_number=part_number,
            )

    async def multipart_initiate_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        content_type: str,
        acl: str = "storageAdmin",
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "key": bucket_key,
        #         "content_type": content_type,
        #         "acl": acl,
        #     }
        # )
        async with gcp_pool.gcp_client() as gcp:
            return await gcp.storage_client().multipart_initiate_upload(
                bucket_name, bucket_key, content_type
            )

    async def multipart_upload_part(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
        data: bytes,
    ) -> str:
        async with gcp_pool.gcp_client() as gcp:
            return await gcp.storage_client().multipart_upload_part(
                bucket_name, bucket_key, upload_id, part_number, data
            )

    async def multipart_complete_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        parts: list[CompletedPartTypeDef],
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "bucket_key": bucket_key,
        #     }
        # )
        async with gcp_pool.gcp_client() as gcp:
            return await gcp.storage_client().multipart_complete_upload(
                bucket_name, bucket_key, upload_id, parts
            )



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, TypedDict
from urllib.parse import quote, urljoin, urlparse
from xml.etree import ElementTree as ET

import aiohttp
import xmltodict
from google.api_core.exceptions import (
    BadRequest,
    Forbidden,
    NotFound,
    ServerError,
    TooManyRequests,
    Unauthorized,
)
from google.auth import default
from google.auth.transport import requests
from google.cloud import storage
from latch_o11y.o11y import dict_to_attrs
from opentelemetry.trace import SpanKind, get_tracer

max_presigned_url_age = timedelta(days=7) // timedelta(seconds=1)

tracer = get_tracer(__name__)


@dataclass(frozen=True)
class Bucket:
    name: str
    id: str
    creation_time: str
    update_time: str
    project_number: str


@dataclass(kw_only=True)
class Object:
    id: str
    name: str
    bucket: str
    size: int
    media_link: str | None
    generation: str | None
    content_type: str | None
    creation_time: datetime | None
    update_time: datetime | None


CompletedPartTypeDef = TypedDict(
    "CompletedPartTypeDef",
    {
        "ETag": str,
        "ChecksumCRC32": str,
        "ChecksumCRC32C": str,
        "ChecksumSHA1": str,
        "ChecksumSHA256": str,
        "PartNumber": int,
    },
    total=False,
)


class AsyncStorageClient:
    credentials, _ = default(scopes=["https://www.googleapis.com/auth/cloud-platform"])

    def __init__(self, session: aiohttp.ClientSession):
        self.api_url = "https://storage.googleapis.com/storage/v1/"
        self.upload_api_url = "https://storage.googleapis.com/upload/storage/v1/"
        self.session = session

    @staticmethod
    def get_auth_token():
        credentials = AsyncStorageClient.credentials

        if credentials.token is None or credentials.expired:
            credentials.refresh(requests.Request())
        return credentials.token

    async def _make_api_request(
        self,
        method: str,  # todo(taras): make an enum
        url: str,
        operation_name: str,
        *,
        data: Any | None = None,
        headers: dict | None = None,
        params: dict | None = None,
        return_bytes=False,
        return_json=False,
        return_response=False,
    ) -> Any:
        endpoint_parts = urlparse(url)

        port = endpoint_parts.port
        if port is None:
            if endpoint_parts.scheme == "http":
                port = 80
            elif endpoint_parts.scheme == "https":
                port = 443

        with tracer.start_as_current_span(
            f"gcp-api.{operation_name}",
            # todo(maximsmol): remove if not needed
            # f"{endpoint_name}.command", # dd span name (kept around in case dd won't recognize the otel one)
            kind=SpanKind.CLIENT,
            attributes={
                # todo(maximsmol): is this service name override a datadog idiosyncrasy?
                "service.name": f"gcp.{operation_name}",
                "http.scheme": endpoint_parts.scheme,
                "http.host": endpoint_parts.netloc,
                "http.endpoint": endpoint_parts.path,
                "net.peer.name": str(endpoint_parts.hostname),
                "rpc.method": operation_name,
                "rpc.system": "gcp-api",
                # datadog:
                "span.type": "http",
                "resource.name": operation_name,
            }
            | (
                {
                    "net.peer.port": port,
                }
                if port is not None
                else {}
            )
            | dict_to_attrs(params if params is not None else {}, "rpc.gcp-api.params"),
        ) as s:

            if headers is None:
                headers = {}

            headers["Authorization"] = f"Bearer {AsyncStorageClient.get_auth_token()}"

            async with self.session.request(
                method,
                url,
                headers=headers,
                params=params,
                data=data,
            ) as resp:
                if resp.status >= 400:
                    err = await resp.text()
                    if resp.status == 400:
                        raise BadRequest(message=err, response=resp)
                    if resp.status == 401:
                        raise Unauthorized(message=err, response=resp)
                    if resp.status == 403:
                        raise Forbidden(message=err, response=resp)
                    if resp.status == 404:
                        raise NotFound(message=err, response=resp)
                    if resp.status == 429:
                        raise TooManyRequests(message=err, response=resp)
                    raise ServerError(message=err, response=resp)
                if resp.status == 204:
                    return
                if return_bytes:
                    return await resp.content.read()
                if return_json:
                    return await resp.json()
                if return_response:
                    return resp
                return await resp.text()

    async def get_bucket(self, bucket_name: str):
        with tracer.start_as_current_span("get_bucket") as s:
            s.set_attributes({"bucket_name": bucket_name})
            try:
                res = await self._make_api_request(
                    "GET",
                    urljoin(
                        self.api_url,
                        f"b/{bucket_name}",
                    ),
                    "GetBucket",
                    return_json=True,
                )
            except NotFound:
                return None

            return Bucket(
                name=res["name"],
                id=res["id"],
                creation_time=res["timeCreated"],
                update_time=res["updated"],
                project_number=res["projectNumber"],
            )

    async def get_blob_meta(self, bucket_name: str, key: str):
        try:
            res = await self._make_api_request(
                "GET",
                urljoin(self.api_url, f"b/{bucket_name}/o/{quote(key, safe='')}"),
                "GetBlobMeta",
                return_json=True,
            )
        except NotFound:
            return None

        return Object(
            name=res["name"],
            id=res["id"],
            bucket=res["bucket"],
            media_link=res["mediaLink"],
            generation=res["generation"],
            content_type=res["contentType"],
            size=int(res["size"]),
            creation_time=datetime.fromisoformat(res["timeCreated"]),
            update_time=datetime.fromisoformat(res["updated"]),
        )

    def list_blobs(
        self,
        bucket_name: str,
        prefix: str | None = None,
        delimiter: str = "/",
        match_glob: str | None = None,
        page_token: str | None = None,
        include_trailing_delimeter: bool | None = None,
        include_folders_as_prefixes: bool | None = None,
        max_items: int | None = None,
    ):
        if match_glob is not None and delimiter != "/":
            raise ValueError("match_glob is only supported with delimeter='/'")

        if include_folders_as_prefixes and delimiter != "/":
            raise ValueError(
                "include_folders_as_prefixes is only supported with delimeter='/'"
            )

        params = {
            "bucket": bucket_name,
            "prefix": prefix,
            "delimiter": delimiter,
            "matchGlob": match_glob,
            "maxResults": 1000,
            "pageToken": page_token,
            "includeTrailingDelimeter": include_trailing_delimeter,
            "includeFoldersAsPrefixes": include_folders_as_prefixes,
        }

        params = {k: str(v) for k, v in params.items() if v is not None}

        return AsyncHTTPIterator(
            url=urljoin(self.api_url, f"b/{bucket_name}/o"),
            headers={},
            params=params,
            storage_client=self,
            max_items=max_items,
        )

    async def get_blob_bytes(
        self,
        bucket_name: str,
        key: str,
        start: int | None = None,
        end: int | None = None,
    ):
        with tracer.start_as_current_span("get_blob_bytes") as s:
            headers = {}
            if start is None:
                start = 0

            if end is not None:
                headers["Range"] = f"bytes={start}-{end}"

            s.set_attributes(
                {
                    "bucket_name": bucket_name,
                    "key": key,
                    "range": headers.get("Range", ""),
                }
            )

            return bytes(
                await self._make_api_request(
                    "GET",
                    urljoin(self.api_url, f"b/{bucket_name}/o/{quote(key, safe='')}"),
                    "GetBlobBytes",
                    params={"alt": "media"},
                    headers=headers,
                    return_bytes=True,
                )
            )

    async def put_blob(
        self,
        bucket_name: str,
        key: str,
        data: bytes,
        content_type: str | None = None,
    ):
        if content_type is None:
            content_type = "text/plain"

        resp = await self._make_api_request(
            "POST",
            urljoin(self.upload_api_url, f"b/{bucket_name}/o"),
            "PutBlob",
            params={"name": key, "uploadType": "media"},
            headers={"Content-Type": content_type},
            data=data,
            return_json=True,
        )

        return Object(
            id=resp["id"],
            name=resp["name"],
            bucket=resp["bucket"],
            media_link=resp["mediaLink"],
            generation=resp["generation"],
            content_type=resp["contentType"],
            size=int(resp["size"]),
            creation_time=datetime.fromisoformat(resp["timeCreated"]),
            update_time=datetime.fromisoformat(resp["updated"]),
        )

    async def copy_blob(
        self,
        src_bucket_name: str,
        src_bucket_key: str,
        dest_bucket_name: str,
        dest_bucket_key: str,
    ):
        done = False
        rewrite_token = None
        try:
            while not done:
                params = {}
                if rewrite_token:
                    params = {"rewriteToken": rewrite_token}

                resp = await self._make_api_request(
                    "POST",
                    urljoin(
                        self.api_url,
                        f"b/{src_bucket_name}/o/{quote(src_bucket_key,safe='')}/rewriteTo/b/{dest_bucket_name}/o/{quote(dest_bucket_key, safe='')}",
                    ),
                    "CopyBlob",
                    params=params,
                    headers={"Content-Type": "application/json"},
                    return_json=True,
                )

                done = resp["done"]
                rewrite_token = resp.get("rewriteToken", None)
        except NotFound:
            return None

    async def delete_blob(
        self, bucket_name: str, key: str, generation: str | None = None
    ):
        params = {}
        if generation is not None:
            params["generation"] = generation

        await self._make_api_request(
            "DELETE",
            urljoin(self.api_url, f"b/{bucket_name}/o/{quote(key, safe='')}"),
            "DeleteBlob",
            params=params,
        )

    def get_signed_download_url(
        self,
        bucket_name: str,
        key: str | None,
        content_disposition: str | None = None,
        content_type: str | None = None,
    ):
        with tracer.start_as_current_span("get_signed_download_url") as s:
            s.set_attributes(
                {
                    "bucket_name": bucket_name,
                    "key": str(key),
                    "content_disposition": str(content_disposition),
                    "content_type": str(content_type),
                }
            )
            # note(taras): this constructs a blob object and signs download url without making requests
            blob = storage.Client().bucket(bucket_name).blob(key)

            return blob.generate_signed_url(
                version="v4",
                expiration=max_presigned_url_age,
                method="GET",
                content_type=content_type,
                response_disposition=content_disposition,
            )

    async def get_signed_part_upload_url(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
    ):
        with tracer.start_as_current_span("get_signed_part_upload_url") as s:
            s.set_attributes(
                {
                    "bucket_name": bucket_name,
                    "bucket_key": bucket_key,
                    "upload_id": upload_id,
                    "part_number": part_number,
                }
            )

            # note(taras): this constructs a blob object and signs
            # upload url without making syncronous requests
            blob = storage.Client().bucket(bucket_name).blob(bucket_key)

            return blob.generate_signed_url(
                version="v4",
                expiration=max_presigned_url_age,
                method="PUT",
                query_parameters={
                    "partNumber": part_number,
                    "uploadId": upload_id,
                },
            )

    async def multipart_initiate_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        content_type: str = "text/plain",
    ):
        meta = await self.get_blob_meta(bucket_name, bucket_key)
        if meta is None:
            await self.put_blob(bucket_name, bucket_key, b"")

        # note(taras): multipart uploads use XML api instead of JSON
        resp = await self._make_api_request(
            "POST",
            f"https://storage.googleapis.com/{bucket_name}/{bucket_key}?uploads",
            "InitiateMultipartUpload",
            headers={
                "Content-Type": content_type,
            },
        )

        return str(xmltodict.parse(resp)["InitiateMultipartUploadResult"]["UploadId"])

    async def multipart_upload_part(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
        data: bytes,
    ) -> str:
        resp = await self._make_api_request(
            "PUT",
            f"https://storage.googleapis.com/{bucket_name}/{bucket_key}?uploadId={upload_id}&partNumber={part_number}",
            "UploadPart",
            data=data,
            return_response=True,
        )

        return resp.headers["ETag"]

    async def multipart_complete_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        parts: list[CompletedPartTypeDef],
    ) -> str:
        root = ET.Element("CompleteMultipartUpload")

        for part in parts:
            if "PartNumber" not in part:
                raise ValueError("PartNumber is required")
            if "ETag" not in part:
                raise ValueError("ETag is required")

            part_element = ET.SubElement(root, "Part")
            ET.SubElement(part_element, "PartNumber").text = str(part["PartNumber"])
            ET.SubElement(part_element, "ETag").text = part["ETag"]

        await self._make_api_request(
            "POST",
            f"https://storage.googleapis.com/{bucket_name}/{bucket_key}?uploadId={upload_id}",
            "CompleteMultipartUpload",
            data=ET.tostring(root),
        )

        blob = await self.get_blob_meta(bucket_name, bucket_key)
        if blob is None:
            raise ServerError("Failed to complete multipart upload: can't find blob")
        if blob.generation is None:
            raise ServerError("The uploaded blob has no generation")

        return blob.generation


class AsyncHTTPIterator:
    def __init__(
        self,
        url: str,
        headers: dict,
        params: dict,
        storage_client: AsyncStorageClient,
        max_items: int | None = None,
    ):
        self.url = url
        self.headers = headers
        self.params = params
        self.next_page_token = None
        self.prefixes: set[str] = set()
        self.items = []
        self.storage_client = storage_client

        self.cur_item = 0
        self.max_items = max_items

    def __aiter__(self):
        return self

    async def __anext__(self):
        if (self.max_items is not None and self.cur_item >= self.max_items) or (
            len(self.items) == 0 and self.next_page_token == ""
        ):
            raise StopAsyncIteration

        if len(self.items) > 0:
            obj = self.items.pop()
            self.cur_item = self.cur_item + 1

            creation_time = obj.get("timeCreated")
            if creation_time is not None:
                creation_time = datetime.fromisoformat(creation_time)

            update_time = obj.get("updated")
            if update_time is not None:
                update_time = datetime.fromisoformat(update_time)

            return Object(
                id=obj["id"],
                name=obj["name"],
                bucket=obj["bucket"],
                media_link=obj.get("mediaLink"),
                generation=obj.get("generation"),
                content_type=obj["contentType"],
                size=int(obj["size"]),
                creation_time=creation_time,
                update_time=update_time,
            )

        if self.next_page_token:
            self.params["pageToken"] = self.next_page_token

        res = await self.storage_client._make_api_request(
            "GET",
            self.url,
            "ListBlobs",
            headers=self.headers,
            params=self.params,
            return_json=True,
        )

        self.items = res.get("items", [])
        prefixes = res.get("prefixes", [])
        self.next_page_token = res.get("nextPageToken", "")

        for prefix in prefixes:
            self.prefixes.add(prefix)

        if "items" not in res:
            raise StopAsyncIteration

        return await self.__anext__()



import asyncio
import math
from collections.abc import AsyncGenerator, Coroutine
from datetime import datetime
from typing import Any

import botocore.exceptions
from latch_aws.aws import max_presigned_url_age
from latch_aws.client_pool import s3_pool

# from latch_o11y.o11y import trace_function_with_span
# from opentelemetry.trace import Span, get_tracer
from types_aiobotocore_s3.type_defs import (
    CompletedPartTypeDef,
)

from ..utils import MP_CHUNK_SIZE, BlobMeta, EmptyMountResponseError, StorageClient


class S3StorageClient(StorageClient):
    async def head(self, bucket_name: str, key: str) -> BlobMeta:
        # s.set_attributes({"bucket_name": bucket_name, "key": key})
        blob = None
        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            try:
                obj = await s3.head_object(Bucket=bucket_name, Key=key)
                if key.endswith("/"):
                    obj["ContentType"] = "dir"
                    obj["ContentLength"] = 0
            except botocore.exceptions.ClientError as e:
                # s.record_exception(e)
                res = e.response
                if res.get("Error", {}).get("Code") != "404":
                    raise e

                try:
                    res = await s3.list_objects_v2(
                        Bucket=bucket_name,
                        Prefix=key,
                        MaxKeys=1,
                    )
                    if "Contents" not in res:
                        raise EmptyMountResponseError()

                    key = key.rstrip("/") + "/"
                    obj = {
                        "ContentType": "dir",
                        "ContentLength": 0,
                        "LastModified": None,
                        "VersionId": None,
                    }
                except botocore.exceptions.ClientError as e:
                    # s.record_exception(e)
                    raise EmptyMountResponseError from e

            blob = BlobMeta(
                bucket=bucket_name,
                key=key,
                content_type=obj["ContentType"],
                size=obj["ContentLength"],
                version=obj.get("VersionId"),
                update_time=obj.get("LastModified"),
            )
        return blob

    async def get_blob_bytes(
        self,
        bucket_name: str,
        key: str,
        start: int | None = None,
        end: int | None = None,
    ) -> bytes:
        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            meta = await self.head(bucket_name=bucket_name, key=key)

            start = start if start is not None else 0
            end = end if end is not None else meta.size - 1

            response = await s3.get_object(
                Bucket=bucket_name,
                Key=key,
                Range=f"bytes={start}-{end}",
            )

            return await response["Body"].read()

    async def list_keys(
        self,
        bucket_name: str,
        prefix: str | None = None,
        delimiter: str = "/",
    ) -> AsyncGenerator[str, Any]:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "prefix": str(prefix),
        #         "delimeter": delimiter,
        #         "page_size": page_size,
        #     }
        # )

        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            paginator = s3.get_paginator("list_objects_v2")

            params = {
                "Bucket": bucket_name,
                "Delimiter": delimiter,
            }
            if prefix is not None:
                params["Prefix"] = prefix

            async for page in paginator.paginate(**params):
                for x in [y["Key"] for y in page.get("Contents", [])] + [
                    y["Prefix"] for y in page.get("CommonPrefixes", [])
                ]:
                    if prefix is None or x.rstrip("/") != prefix.rstrip("/"):
                        yield x

    async def put_blob(
        self,
        bucket_name: str,
        key: str,
        body: bytes = b"",
        content_type: str = "text/plain",
        acl: str = "bucket-owner-full-control",
    ) -> BlobMeta:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "key": key,
        #         "body": body,
        #         "content_type": str(content_type),
        #     }
        # )

        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            res = await s3.put_object(
                Bucket=bucket_name,
                Key=key,
                Body=body,
                ContentType=content_type,
                ACL=acl,
            )

            return BlobMeta(
                bucket=bucket_name,
                key=key,
                content_type=content_type,
                size=len(body),
                version=res["VersionId"],
                update_time=datetime.now(),
            )

    async def copy_blob(
        self,
        src_key: str,
        src_bucket: str,
        dest_key: str,
        dest_bucket: str,
    ) -> None:
        # s.set_attributes(
        #     {
        #         "src_key": src_key,
        #         "src_bucket": src_bucket,
        #         "dest_key": dest_key,
        #         "dest_bucket": dest_bucket,
        #     }
        # )

        blob = await self.head(src_bucket, src_key)
        if blob is None:
            return
        async with s3_pool.s3_client_for_bucket(dest_bucket) as s3:
            await s3.copy_object(
                CopySource={
                    "Bucket": src_bucket,
                    "Key": src_key,
                },
                Bucket=dest_bucket,
                Key=dest_key,
                ContentType=blob.content_type,
                ACL="bucket-owner-full-control",
            )

    async def copy_blob_multipart(
        self,
        src_key: str,
        src_bucket: str,
        dest_key: str,
        dest_bucket: str,
    ) -> None:
        # s.set_attributes(
        #     {
        #         "src_key": src_key,
        #         "src_bucket": src_bucket,
        #         "dest_key": dest_key,
        #         "dest_bucket": dest_bucket,
        #     }
        # )

        blob = await self.head(src_bucket, src_key)
        if blob is None:
            raise EmptyMountResponseError()

        async with s3_pool.s3_client_for_bucket(dest_bucket) as s3:
            upload_id = await self.multipart_initiate_upload(
                dest_bucket,
                dest_key,
                blob.content_type,
                "bucket-owner-full-control",
            )

            parts = math.ceil(blob.size / MP_CHUNK_SIZE)
            # s.set_attribute("copy.nrof_parts", parts)

            async def run_upload_chunk(
                byte_range: str, index: int
            ) -> CompletedPartTypeDef:
                res = await s3.upload_part_copy(
                    CopySource={
                        "Bucket": src_bucket,
                        "Key": src_key,
                    },
                    Bucket=dest_bucket,
                    Key=dest_key,
                    PartNumber=index + 1,
                    UploadId=upload_id,
                    CopySourceRange=byte_range,
                )
                if "CopyPartResult" not in res or "ETag" not in res["CopyPartResult"]:
                    raise ValueError("Etag not in response")

                return {
                    "ETag": res["CopyPartResult"]["ETag"],
                    "PartNumber": index + 1,
                }

            upload_chunk_requests: list[Coroutine[Any, Any, CompletedPartTypeDef]] = []

            upload_chunk_responses: list[CompletedPartTypeDef] = []
            for i in range(parts):
                byte_start = i * MP_CHUNK_SIZE
                byte_end = (i + 1) * MP_CHUNK_SIZE - 1
                byte_end = min(byte_end, blob.size - 1)
                byte_range = f"bytes={byte_start}-{byte_end}"

                upload_chunk_requests.append(run_upload_chunk(byte_range, i))

                if len(upload_chunk_requests) >= 50:
                    upload_chunk_responses.extend(
                        await asyncio.gather(*upload_chunk_requests)
                    )
                    upload_chunk_requests = []

            upload_chunk_responses.extend(await asyncio.gather(*upload_chunk_requests))

            await self.multipart_complete_upload(
                dest_bucket, dest_key, upload_id, upload_chunk_responses
            )

    async def delete_blob(
        self,
        bucket_name: str,
        key: str,
    ) -> None:
        # s.set_attributes({"bucket_name": bucket_name, "key": key})
        try:
            async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
                await s3.delete_object(
                    Bucket=bucket_name,
                    Key=key,
                )
        except botocore.exceptions.ClientError as e:
            res = e.response
            if res.get("Error", {}).get("Code") != "404":
                raise e

            return

    async def generate_signed_download_url(
        self,
        bucket_name: str,
        key: str | None,
        content_disposition: str | None = None,
        content_type: str | None = None,
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "key": str(key),
        #         "content_disposition": str(content_disposition),
        #         "content_type": str(content_type),
        #     }
        # )

        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            return await s3.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": bucket_name,
                    "Key": key,
                    # todo(maximsmol): add extra validation
                    # "ExpectedBucketOwner": "",
                    # "IfMatch": "",
                    # "VersionId": ""
                }
                | (
                    {"ResponseContentDisposition": content_disposition}
                    if content_disposition is not None
                    else {}
                )
                | (
                    {"ResponseContentType": content_type}
                    if content_type is not None
                    else {}
                ),
                ExpiresIn=max_presigned_url_age,
            )

    async def generate_signed_part_upload_url(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "bucket_key": bucket_key,
        #         "part_number": part_number,
        #     }
        # )

        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            return await s3.generate_presigned_url(
                "upload_part",
                Params={
                    "Bucket": bucket_name,
                    "Key": bucket_key,
                    "UploadId": upload_id,
                    "PartNumber": part_number,
                    # todo(maximsmol): add extra validation
                    # "ExpectedBucketOwner": "",
                },
                ExpiresIn=max_presigned_url_age,
            )

    async def multipart_initiate_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        content_type: str,
        acl: str = "bucket-owner-full-control",
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "bucket_key": bucket_key,
        #         "content_type": content_type,
        #         "acl": acl,
        #     }
        # )
        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            multipart_res = await s3.create_multipart_upload(
                Bucket=bucket_name,
                Key=bucket_key,
                ACL=acl,
                ContentType=content_type,
                # todo(maximsmol): add extra validation
                # "ExpectedBucketOwner": "",
            )
            return multipart_res["UploadId"]

    async def multipart_upload_part(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
        data: bytes,
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "bucket_key": bucket_key,
        #         "part_number": part_number,
        #     }
        # )

        async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
            ret = await s3.upload_part(
                Bucket=bucket_name,
                Key=bucket_key,
                UploadId=upload_id,
                PartNumber=part_number,
                Body=data,
            )
            return ret["ETag"]

    async def multipart_complete_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        parts: list[CompletedPartTypeDef],
    ) -> str:
        # s.set_attributes(
        #     {
        #         "bucket_name": bucket_name,
        #         "bucket_key": bucket_key,
        #         "upload_id": upload_id,
        #     }
        # )

        try:
            async with s3_pool.s3_client_for_bucket(bucket_name) as s3:
                res = await s3.complete_multipart_upload(
                    Bucket=bucket_name,
                    Key=bucket_key,
                    UploadId=upload_id,
                    MultipartUpload={"Parts": parts},
                    # todo(maximsmol): add extra validation
                    # "ExpectedBucketOwner": "",
                )

                return res["VersionId"]
        except botocore.exceptions.ClientError as e:
            err_res = e.response
            if "Error" not in err_res:
                raise e

            if "Code" not in err_res["Error"]:
                raise e

            if err_res["Error"]["Code"] != "EntityTooSmall":
                raise e

            raise ValueError("Upload size is less than minimum allowed") from e



from asyncio import gather
from asyncio.locks import BoundedSemaphore
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass

import aiohttp
from opentelemetry.trace import get_tracer

from .storage import AsyncStorageClient

tracer = get_tracer(__name__)


class AsyncGCPClient:
    def __init__(self, session: aiohttp.ClientSession):
        self.session = session

    def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.session.__aexit__(exc_type, exc_val, exc_tb)

    def storage_client(self):
        return AsyncStorageClient(self.session)


@dataclass
class ClientList:
    sema: BoundedSemaphore
    client_add_sema: BoundedSemaphore
    clients: list[AsyncGCPClient]


class AsyncGCPClientPool:
    max_clients: int | None = None
    client_list: ClientList | None = None

    async def open(self, max_clients: int):
        with tracer.start_as_current_span(
            "open pool", attributes={"max_clients": max_clients}
        ):
            self.max_clients = max_clients
            self.client_list = ClientList(
                sema=BoundedSemaphore(self.max_clients),
                client_add_sema=BoundedSemaphore(1),
                clients=[],
            )

    async def close(self):
        with tracer.start_as_current_span("close pool"):
            if self.max_clients is None or self.client_list is None:
                raise RuntimeError("Pool not open")

            close_tasks = []
            for client in self.client_list.clients:
                close_tasks.append(client.__aexit__(None, None, None))

            await gather(*close_tasks)
            self.client_list = None
            self.max_clients = None

    @asynccontextmanager
    async def gcp_client(self) -> AsyncGenerator[AsyncGCPClient, None]:
        with tracer.start_as_current_span("get client") as s:
            if self.max_clients is None or self.client_list is None:
                raise RuntimeError("Pool not open")

            async with self.client_list.sema:
                async with self.client_list.client_add_sema:
                    if len(self.client_list.clients) == 0:
                        with tracer.start_as_current_span(
                            "create gcp client",
                        ):
                            http_session = await aiohttp.ClientSession().__aenter__()
                            client = AsyncGCPClient(http_session)
                    else:
                        client = self.client_list.clients.pop()

                try:
                    yield client
                finally:
                    self.client_list.clients.append(client)
                    s.set_attributes(
                        {
                            "client_count": len(self.client_list.clients),
                        }
                    )


gcp_pool = AsyncGCPClientPool()



from enum import Enum

from .aws.storage_client import S3StorageClient
from .gcp.storage_client import GCPStorageClient
from .utils import StorageClient


class LDataNodeType(str, Enum):
    account_root = "account_root"
    dir = "dir"
    obj = "obj"
    mount = "mount"
    link = "link"
    mount_gcp = "mount_gcp"
    mount_azure = "mount_azure"


def get_storage_client(root_type: LDataNodeType | None) -> StorageClient:
    if root_type is not None and root_type not in [
        LDataNodeType.account_root,
        LDataNodeType.mount,
        LDataNodeType.mount_gcp,
        LDataNodeType.mount_azure,
    ]:
        raise ValueError("Root is not a mount")

    if root_type == LDataNodeType.mount_gcp:
        return GCPStorageClient()
    else:
        return S3StorageClient()
